import copy
import inspect
import re
import typing as t
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field

import click as _click

from .types import ClassUnion, List, Tuple
from .utils import (
    _full_signature,
    _get_help_text,
    _is_class,
    _wrap_fn_to_allow_kwargs_instead_of_args,
    get_class_name,
    Literal,
    SignatureWithDescription,
)


_SUPPORTED_TYPES = [str, float, int, bool]


@dataclass
class AClickParameter:
    parameter_type: t.Type
    parameter_name: str
    parameter_default: t.Any = None
    children: t.List["AClickParameter"] = field(default_factory=list)
    is_hierarchical: bool = False
    positional_only: bool = False
    keyword_only: bool = False
    expanded_parameter_type: t.Optional[t.Type] = None
    description: t.Optional[str] = None

    @property
    def name(self):
        return self.parameter_name


class ParameterRenamer(metaclass=ABCMeta):
    """A base class used as parameter renamer for a :class:`Command`.
    The derived class must implement the call method that
    maps from the original name to a new name.
    """

    @abstractmethod
    def __call__(self, name: str) -> str:
        """Maps the parameter name into a new name.

        :param name: Original name of the parameter.
        :return: The new name of the parameter.
        """
        raise NotImplementedError()  # pragma: no cover


@dataclass
class AClickContext:
    params: t.Optional[t.List[AClickParameter]] = None


class Command(_click.Command):
    """
    Command class extend `click.Command` class and it automatically
    parses callback's signature and generates the parameters.
    It also supports dynamic parsing, where based on some
    parameter, other parameters may be added, which is used in the
    hierarchical parsing (see `is_hierarchical`).

    :param name: the name of the command to use unless a group overrides it.
    :param context_settings: an optional dictionary with defaults that are
                             passed to the context object.
    :param callback: the callback to invoke.  This is optional.
    :param signature: signature to be used for automatic parameter registration.
                      If no signature is specified, it is parsed automatically
                      from the callback.
    :param hierarchical: whether to parse complex types using hierarchical
                         parsing instead of inline parsing. In hierarchical
                         parsing the complex class options are expanded into
                         individual options corresponding to individual
                         properties. Note that some types are not supported
                         with hierarchical parsing. Default is to use inline
                         parsing.
    :param map_parameter_name: a function that maps from parameter paths
                               (e.g. param1.property1.subproperty2) into
                               parameter name used in parsing.
    :param params: the parameters to register with this command. This can
                   be either :class:`Option` or :class:`Argument` objects.
                   These parameters override automatically generated params
                   if they share the same name.
    :param help: the help string to use for this command. The callback's
                 docstring is parsed if no help is passed.
    :param epilog: like the help string but it's printed at the end of the
                   help page after everything else.
    :param short_help: the short help to use for this command.  This is
                       shown on the command listing of the parent command.
                       The callback's docstring is parsed if no help is passed.
    :param add_help_option: by default each command registers a ``--help``
                            option.  This can be disabled by this parameter.
    :param no_args_is_help: this controls what happens if no arguments are
                            provided.  This option is disabled by default.
                            If enabled this will add ``--help`` as argument
                            if no arguments are passed
    :param hidden: hide this command from help outputs.
    :param deprecated: issues a message indicating that
                             the command is deprecated.
    """

    def __init__(
        self,
        *args,
        signature: t.Optional[inspect.Signature] = None,
        hierarchical: bool = False,
        map_parameter_name: t.Optional[ParameterRenamer] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if signature is None:
            if self.callback is None:
                raise ValueError("Either signature or callback must be specified")
            signature = _full_signature(self.callback)
            self.callback = _wrap_fn_to_allow_kwargs_instead_of_args(self.callback)
        self.callback_signature: inspect.Signature = signature
        self.hierarchical = hierarchical
        self.map_parameter_name = map_parameter_name
        self._assert_signature_is_supported(self.callback_signature)
        if (self.help is None or self.short_help is None) and isinstance(
            signature, SignatureWithDescription
        ):
            short_help, help = _get_help_text(signature)
            self.short_help = (
                self.short_help if self.short_help is not None else short_help
            )
            self.help = self.help if self.help is not None else help

    def _wrap_context(self, ctx) -> t.Tuple[_click.Context, AClickContext]:
        aclick_context = getattr(ctx, "__aclick_context__", None)
        if aclick_context is None:
            aclick_context = AClickContext()
            setattr(ctx, "__aclick_context__", aclick_context)
        return ctx, aclick_context

    def add_conditional_params(
        self, ctx: _click.Context, aclick_ctx: AClickContext
    ) -> bool:
        params_added = False

        def walk(path, p):
            nonlocal params_added
            name = "_".join(path + [p.parameter_name]).replace("-", "_")
            if (
                self.hierarchical
                and p.is_hierarchical
                and getattr(p.parameter_type, "__origin__", None) is t.Union
                and p.expanded_parameter_type is None
            ):
                # Switch the type if not already expanded
                if name in ctx.params:
                    class_name = ctx.params.get(name)
                    supported_classes = {
                        x for x in p.parameter_type.__args__ if _is_class(x)
                    }
                    parameter_type = next(
                        (
                            x
                            for x in supported_classes
                            if get_class_name(x) == class_name
                        ),
                        None,
                    )
                    if parameter_type is None:
                        supported_classes_names = ", ".join(
                            sorted(map(get_class_name, supported_classes))
                        )
                        raise _click.ClickException(
                            f'Class with name "{class_name}" was not found in the set of supported classes {{{supported_classes_names}}}'
                        )
                    p.expanded_parameter_type = parameter_type
                    p.children = _parse_signature(
                        _full_signature(parameter_type), hierarchical=self.hierarchical
                    )
                    params_added = True

            for pc in p.children:
                walk(path + [p.parameter_name], pc)

        if aclick_ctx.params is not None:
            for p in aclick_ctx.params:
                walk([], p)
        return params_added

    def build_click_parameter(
        self,
        option_name: t.Sequence[str],
        parameter_type: t.Type,
        default: t.Any,
        help: t.Optional[str],
        is_argument: bool,
    ) -> t.Optional[_click.Parameter]:
        kwargs = dict()
        if default is not _empty:
            kwargs["default"] = default
            kwargs["required"] = False
        else:
            kwargs["required"] = True
        if help is not None and not is_argument:
            kwargs["help"] = help
        original_parameter_type = parameter_type
        cls = _click.Argument if is_argument else _click.Option

        # Handle optional
        if (
            getattr(parameter_type, "__origin__", None) is t.Union
            and len(parameter_type.__args__) == 2
            and type(None) in parameter_type.__args__
        ):
            parameter_type = next(
                x for x in parameter_type.__args__ if x not in (type(None),)
            )

        if _is_class(parameter_type):
            kwargs["type"] = ClassUnion([parameter_type])
            return cls(option_name, **kwargs)

        if getattr(parameter_type, "__origin__", None) is Literal:
            values = parameter_type.__args__
            if all(isinstance(x, str) for x in values):
                kwargs["type"] = _click.types.Choice(values, case_sensitive=False)
                return cls(option_name, **kwargs)

        if getattr(parameter_type, "__origin__", None) is t.Union:
            # Handle union type
            nonnull_args = [
                x for x in parameter_type.__args__ if x not in (type(None),)
            ]
            all_classes = all(map(_is_class, nonnull_args))
            if not all_classes:
                raise ValueError(
                    f"Parameter with parameter name {option_name[-1]} of type {original_parameter_type} is not supported. "
                    "Only unions of all classes are supported."
                )

            types = [x for x in parameter_type.__args__ if _is_class(x)]
            kwargs["type"] = ClassUnion(types)
            return cls(option_name, **kwargs)

        if getattr(parameter_type, "__origin__", None) is list:
            kwargs["type"] = List(parameter_type.__args__[0])
            return cls(option_name, **kwargs)
        if getattr(parameter_type, "__origin__", None) is tuple:
            kwargs["type"] = Tuple(list(parameter_type.__args__))
            return cls(option_name, **kwargs)
        if parameter_type in _SUPPORTED_TYPES:
            kwargs["type"] = parameter_type
            if parameter_type is bool and not is_argument:
                kwargs["type"] = _click.types.BoolParamType()
                option_name = [
                    f"{x}/--no-{x[2:]}" if x.startswith("--") else x
                    for x in option_name[:-1]
                ] + [option_name[-1]]
                option = _click.Option(option_name, **kwargs)
                if "default" not in kwargs:
                    # NOTE: this is fix for 8.0.x version of click
                    option.default = None
                return option
            else:
                return cls(option_name, **kwargs)
        return None

    def _assert_signature_is_supported(self, signature: inspect.Signature) -> None:
        aclick_params = _parse_signature(signature)

        def walk(path, p):
            output_parameter_name = "_".join(path + [p.parameter_name])
            if p.parameter_type is _empty:
                raise ValueError(
                    f'Parameter named "{output_parameter_name}" does not have it\'s type specified.'
                )

            if p.positional_only:
                if len(path) > 0:
                    raise ValueError(
                        f'Positional arguments are allowed only on top level. Failed argument name is "{output_parameter_name}".'
                    )
                assert not self.hierarchical or not p.is_hierarchical, (
                    f"Hierarchical arguments (name: {output_parameter_name}) are not supported. "
                    "Only simple types can be used in hierarchical parsing"
                )
                parameter = self.build_click_parameter(
                    [output_parameter_name],
                    p.parameter_type,
                    p.parameter_default,
                    help=p.description,
                    is_argument=True,
                )
                if parameter is None:
                    raise ValueError(
                        f"Argument with name {output_parameter_name} of type {p.parameter_type} could not be converted to a click Argument"
                    )

            elif self.hierarchical and p.is_hierarchical:
                # For hierarchical, we support only union of classes or optional
                if getattr(p.parameter_type, "__origin__", None) is t.Union:
                    nonnull_args = [
                        x for x in p.parameter_type.__args__ if x not in (type(None),)
                    ]
                    all_classes = all(map(_is_class, nonnull_args))
                    if not all_classes:
                        raise ValueError(
                            f"Parameter with parameter name {output_parameter_name} of type {p.parameter_type} is not supported. "
                            "Only unions of all classes are supported."
                        )

                    for x in nonnull_args:
                        for pc in _parse_signature(
                            _full_signature(x), hierarchical=self.hierarchical
                        ):
                            walk(path + [p.parameter_name, pc.parameter_name], pc)
                elif _is_class(p.parameter_type):
                    for pc in _parse_signature(_full_signature(p.parameter_type)):
                        walk(path + [p.parameter_name, pc.parameter_name], pc)
            else:
                parameter = self.build_click_parameter(
                    [f"--{output_parameter_name}", output_parameter_name],
                    p.parameter_type,
                    p.parameter_default,
                    help=p.description,
                    is_argument=False,
                )
                if parameter is None:
                    raise ValueError(
                        f"Parameter with parameter name {output_parameter_name} of type {p.parameter_type} could not be converted to a click Option"
                    )

        for p in aclick_params:
            walk([], p)

    def _get_click_parameters(self, parameters: t.List[AClickParameter]):
        def walk(path, p):
            kwargs: t.Dict[str, t.Any] = dict()
            children = p.children
            output_parameter_name = "_".join(path + [p.parameter_name])
            if p.positional_only:
                opt_name = [output_parameter_name]
            else:
                name = "-".join(path + [p.parameter_name]).replace("_", "-")
                name = ".".join(path + [p.parameter_name])
                if self.map_parameter_name is not None:
                    name = self.map_parameter_name(name)
                name = name.replace(".", "-").replace("_", "-")
                opt_name = [f"--{name}", output_parameter_name]
            parameter_type = p.expanded_parameter_type or p.parameter_type

            if not p.positional_only and self.hierarchical and p.is_hierarchical:
                original_parameter_type = p.parameter_type
                if (
                    getattr(original_parameter_type, "__origin__", None) is t.Union
                    and sum(1 for x in original_parameter_type.__args__ if _is_class(x))
                    > 1
                ):
                    values = [
                        get_class_name(x)
                        for x in original_parameter_type.__args__
                        if _is_class(x)
                    ]
                    kwargs["type"] = _click.types.Choice(values, case_sensitive=False)
                    kwargs["is_eager"] = True
                    yield _click.Option(opt_name, **kwargs)
            else:
                parameter = self.build_click_parameter(
                    opt_name,
                    parameter_type,
                    p.parameter_default,
                    help=p.description,
                    is_argument=p.positional_only,
                )
                assert (
                    parameter is not None
                ), f"Parameter with parameter name {output_parameter_name} of type {p.parameter_type} could not be converted to a click Parameter"
                yield parameter

            if not p.positional_only and self.hierarchical and p.is_hierarchical:
                for pc in children:
                    yield from walk(path + [p.parameter_name], pc)

        for p in parameters:
            yield from walk([], p)

    def get_params(self, ctx: _click.Context):
        ctx, aclick_ctx = self._wrap_context(ctx)
        assert (
            aclick_ctx.params is not None
        ), "parse_args method was not called before get_params method"

        aclick_params = aclick_ctx.params
        params = super().get_params(ctx)
        existing_names = {x.name for x in params}
        params.extend(
            x
            for x in self._get_click_parameters(aclick_params)
            if x.name not in existing_names
        )
        return params

    def parse_args(self, ctx: _click.Context, args: t.List[str]) -> t.List[str]:
        original_args = args
        ctx, aclick_ctx = self._wrap_context(ctx)
        aclick_ctx.params = _parse_signature(
            self.callback_signature, hierarchical=self.hierarchical
        )
        if self.hierarchical:
            arguments_added = True
            while arguments_added:
                local_ctx = copy.deepcopy(ctx)
                local_ctx.ignore_unknown_options = True
                arguments_added = False
                parser = self.make_parser(local_ctx)
                opts, args, param_order = parser.parse_args(
                    args=copy.deepcopy(original_args)
                )

                params = self.get_params(local_ctx)
                for param in _click.core.iter_params_for_processing(
                    param_order, params
                ):
                    value, args = param.handle_parse_result(local_ctx, opts, args)

                # Add new params
                arguments_added = self.add_conditional_params(local_ctx, aclick_ctx)
        result = super().parse_args(ctx, original_args)

        # Bind hierarchical parameters to correct types
        if self.hierarchical:
            params = ctx.params
            remove_parameters: t.Set[str] = set()

            def build_obj(path, p, used_parameters):
                name = "_".join(path + [p.parameter_name]).replace("-", "_")
                if not p.is_hierarchical or not self.hierarchical:
                    if name in ctx.params:
                        used_parameters.add(name)
                    return ctx.params.get(name, _empty)

                kwargs = dict()
                for pc in p.children:
                    kwargs[pc.parameter_name] = build_obj(
                        path + [p.parameter_name], pc, used_parameters
                    )
                parameter_type = p.expanded_parameter_type or p.parameter_type
                if _is_optional(parameter_type) and name not in params:
                    # If no class was selected, we return default
                    return (
                        None if p.parameter_default is _empty else p.parameter_default
                    )

                assert (
                    getattr(parameter_type, "__origin__", None) is not t.Union
                ), f'Parameter with name "{name}" was not expanded'
                return parameter_type(**kwargs)

            keep_parameters = set()
            for p in aclick_ctx.params:
                if self.hierarchical and p.is_hierarchical:
                    assert p.parameter_name is not None
                    assert not p.positional_only
                    params[p.parameter_name] = build_obj([], p, remove_parameters)
                    keep_parameters.add(p.name)
            remove_parameters.difference_update(keep_parameters)
            for pr in remove_parameters:
                params.pop(pr)
        return result


class Group(_click.Group, Command):
    @t.overload
    def command(self, __func: t.Callable[..., t.Any]) -> Command:
        ...  # pragma: no cover

    @t.overload
    def command(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Callable[[t.Callable[..., t.Any]], Command]:
        ...  # pragma: no cover

    def command(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Union[t.Callable[[t.Callable[..., t.Any]], Command], Command]:
        """A shortcut decorator for declaring and attaching a command to
        the group. This takes the same arguments as :func:`command` and
        immediately registers the created command with this group by
        calling :meth:`add_command`.
        To customize the command class used, set the
        :attr:`command_class` attribute.
        """
        from .decorators import command

        if self.command_class and kwargs.get("cls") is None:
            kwargs["cls"] = self.command_class

        func: t.Optional[t.Callable] = None

        if args and callable(args[0]):
            assert (
                len(args) == 1 and not kwargs
            ), "Use 'command(**kwargs)(callable)' to provide arguments."
            (func,) = args
            args = ()

        def decorator(f: t.Callable[..., t.Any]) -> Command:
            cmd: Command = command(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        if func is not None:
            return decorator(func)

        return decorator

    @t.overload
    def group(self, __func: t.Callable[..., t.Any]) -> "Group":
        ...  # pragma: no cover

    @t.overload
    def group(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Callable[[t.Callable[..., t.Any]], "Group"]:
        ...  # pragma: no cover

    def group(
        self, *args: t.Any, **kwargs: t.Any
    ) -> t.Union[t.Callable[[t.Callable[..., t.Any]], "Group"], "Group"]:
        """A shortcut decorator for declaring and attaching a group to
        the group. This takes the same arguments as :func:`group` and
        immediately registers the created group with this group by
        calling :meth:`add_command`.
        To customize the group class used, set the :attr:`group_class`
        attribute.
        """
        from .decorators import group

        func: t.Optional[t.Callable] = None

        if args and callable(args[0]):
            assert (
                len(args) == 1 and not kwargs
            ), "Use 'group(**kwargs)(callable)' to provide arguments."
            (func,) = args
            args = ()

        if self.group_class is not None and kwargs.get("cls") is None:
            if self.group_class is type:
                kwargs["cls"] = type(self)
            else:
                kwargs["cls"] = self.group_class

        def decorator(f: t.Callable[..., t.Any]) -> "Group":
            cmd: Group = group(*args, **kwargs)(f)
            self.add_command(cmd)
            return cmd

        if func is not None:
            return decorator(func)

        return decorator


class RegexParameterRenamer:
    """The :class:`RegexParameterRenamer` replaces parameter name
    using regex matching and substitution. The first pattern
    that matches the parameter name is used and the matched part
    of the parameter name is substituted with the replacement value.
    If no regex expression matches the parameter name, the name is
    returned unchanged.

    :attr patterns: List of pairs of patterns and replacement values.
    """

    def __init__(self, patterns: t.List[t.Tuple[str, str]]):
        """The :class:`RegexParameterRenamer` replaces parameter name
        using regex matching and substitution. The first pattern
        that matches the parameter name is used and the matched part
        of the parameter name is substituted with the replacement value.
        If no regex expression matches the parameter name, the name is
        returned unchanged.

        :param patterns: List of pairs of patterns and the associated
                         replacement values.
        """
        self.patterns = patterns
        self._compiled_patterns = [re.compile(x) for x, _ in patterns]
        self._fast_pattern = re.compile(
            "(?:"
            + "|".join(f"(?P<pp{i}>{x})" for i, (x, _) in enumerate(patterns))
            + ")"
        )

    def __call__(self, name: str) -> str:
        """Maps the parameter name into a new name.

        :param name: Original name of the parameter.
        :return: The new name of the parameter.
        """
        found_pattern_match = self._fast_pattern.match(name)
        if not found_pattern_match:
            return name
        found_pattern_dict = found_pattern_match.groupdict()
        found_pattern_index = next(
            i for i in range(len(self.patterns)) if found_pattern_dict[f"pp{i}"]
        )
        found_pattern, replacement = (
            self._compiled_patterns[found_pattern_index],
            self.patterns[found_pattern_index][1],
        )
        return found_pattern.sub(replacement, name)


class FlattenParameterRenamer(RegexParameterRenamer):
    """The :class:`FlattenParameterRenamer` is used to map
    parameters into lower level parameters (e.g. `"class1.prop1"` into `"prop1"`).
    This is especially useful for hierarchical parsing, where properties
    deep in the class structure can have long names.
    The number of levels to remove is a parameter of the class.
    """

    def __init__(self, num_levels: int = -1):
        """The :class:`FlattenParameterRenamer` is used to map
        parameters into lower level parameters (e.g. `"class1.prop1"` into `"prop1"`).
        This is especially useful for hierarchical parsing, where properties
        deep in the class structure can have long names.

        :param num_levels: This parameter specifies how many levels to remove.
                           E.g. with `num_levels=1`, `"class1.a.prop1"` becomes `"a.prop1"`.
                           If the value is `-1` (default), infinitely many levels are removed
                           and only the last part of the parameter path is used.
        """
        self._num_levels = num_levels
        num_levels_str = str(num_levels) if num_levels != -1 else ""
        super().__init__([(fr"(?:[^\.]+\.){{,{num_levels_str}}}(.*)", r"\1")])


class _empty:
    """Marker object for Signature.empty and Parameter.empty."""


def _parse_signature(
    signature: t.Union[inspect.Signature, SignatureWithDescription],
    hierarchical: bool = False,
) -> t.List[AClickParameter]:
    params = []
    for p in signature.parameters.values():
        if (
            p.kind == inspect.Parameter.KEYWORD_ONLY
            or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            or p.kind == inspect.Parameter.POSITIONAL_ONLY
        ):
            children = []
            parameter_type = p.annotation
            parameter_default = _empty if p.default is inspect._empty else p.default
            is_hierarchical = False
            if parameter_type is inspect._empty:
                parameter_type = _empty
            elif p.kind != inspect.Parameter.POSITIONAL_ONLY:
                if _is_class(parameter_type):
                    is_hierarchical = True
                    if hierarchical:
                        child_signature = _full_signature(parameter_type)
                        children = _parse_signature(
                            child_signature, hierarchical=hierarchical
                        )
                elif getattr(parameter_type, "__origin__", None) is t.Union and any(
                    _is_class(x) for x in parameter_type.__args__
                ):
                    is_hierarchical = True
            if is_hierarchical:
                if parameter_default is not _empty and parameter_default is not None:
                    raise ValueError(
                        "Cannot use non-default parameter with hierarchical parsing.\n"
                        "Name of the bad parameter is: {p.name}"
                    )

            param = AClickParameter(
                parameter_name=p.name,
                parameter_type=parameter_type,
                children=children,
                is_hierarchical=is_hierarchical,
                positional_only=(p.kind == inspect.Parameter.POSITIONAL_ONLY),
                keyword_only=(p.kind == inspect.Parameter.KEYWORD_ONLY),
                parameter_default=parameter_default,
                description=getattr(p, "description", None),
            )

            params.append(param)
    return params


def _is_optional(tp):
    return (
        getattr(tp, "__origin__", None) is t.Union
        and sum(1 for x in tp.__args__ if x not in (type(None),)) == 1
    )



import typing as t

import click as _click
from click.types import ParamType

from .utils import _ClassArgument, _is_class, get_class_name, parse_class_structure


class Tuple(_click.Tuple):
    """The default behavior of Click is to apply a type on a value directly.
    This works well in most cases, except for when `nargs` is set to a fixed
    count and different types should be used for different items.  In this
    case the :class:`Tuple` type can be used.  This type can only be used
    if `nargs` is set to a fixed number.
    This can be selected by using a Python tuple literal as a type.

    As opposed to Click's original Tuple type, this type supports inline
    class parsing. A complicated structure of classes can be parsed
    using this type. See :func:`parse_class_structure`.

    :param types: a list of types that should be used for the tuple items.
    """

    def __init__(self, types: t.Sequence[t.Union[t.Type, ParamType]]) -> None:
        super().__init__([convert_type(ty) for ty in types])


class List(_click.ParamType):
    """This type parses a list of values separated by commas as individual
    instances.
    This type supports inline
    class parsing. A complicated structure of classes can be parsed
    using this type. See :func:`parse_class_structure`.


    :param inner_type: Type of the objects in the list.
    """

    name = "list"

    def __init__(self, inner_type: t.Type):
        self.inner_type = convert_type(inner_type)
        self.name = f"list of {self.inner_type.name}s"

    def __repr__(self) -> str:
        return f"LIST[{repr(self.inner_type)}]"

    def convert(
        self,
        value: t.Any,
        param: t.Optional[_click.Parameter],
        ctx: t.Optional[_click.Context],
    ) -> t.Any:
        assert isinstance(value, str)
        values = [str(x) for x in _ClassArgument.from_str(f"list({value})").args]
        return list(map(self.inner_type, values))


class ClassUnion(_click.ParamType):
    """This type parses the string value as the instance of
    one of specified classes. A complicated structure of
    classes can be parsed using this type.
    See :func:`parse_class_structure`.

    :param classes: List of classes to parse into.
    :param known_classes: List of classes supported at deeper levels (e.g. as property values).
    """

    name = "class union"

    def __init__(
        self, classes: t.List[t.Type], known_classes: t.Optional[t.List[t.Type]] = None
    ):
        self.classes = classes
        self.known_classes = known_classes

    def __repr__(self) -> str:
        return "|".join(sorted(get_class_name(x) for x in self.classes))

    def convert(
        self,
        value: t.Any,
        param: t.Optional[_click.Parameter],
        ctx: t.Optional[_click.Context],
    ) -> t.Any:

        assert isinstance(value, str)
        value = parse_class_structure(value, self.classes, self.known_classes)
        return value


def convert_type(ty: t.Optional[t.Any], default: t.Optional[t.Any] = None) -> ParamType:
    if ty is not None and _is_class(ty):
        return ClassUnion([ty])
    if ty is not None and getattr(ty, "__origin__", None) is t.Union:
        classes = [c for c in ty.__args__ if _is_class(c)]
        return ClassUnion(classes)
    if ty is not None and getattr(ty, "__origin__", None) is list:
        return List(ty.__args__[0])
    return _click.types.convert_type(ty, default)



import builtins
import dataclasses
import inspect
import re
import types
import typing as t
from collections import deque, OrderedDict
from functools import partial
from itertools import chain

import docstring_parser

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

T = t.TypeVar("T", bound=t.Callable)
TType = t.TypeVar("TType", bound=t.Type)


class ParameterWithDescription(inspect.Parameter):
    """
    Extends `inspect.Parameter` class with description parameter.

    :attr name: The name of the parameter as a string.
    :attr default: The default value for the parameter if specified, otherwise `inspect._empty`.
    :attr annotation: The annotation for the parameter if specified, otherwise `inspect._empty`.
    :attr kind: Describes how argument values are bound to the parameter.
    :attr description: String description of the parameter.
    """

    def __init__(self, *args, description: t.Optional[str] = None, **kwargs):
        """
        Extends `inspect.Parameter` class with description parameter.

        :param name: The name of the parameter as a string.
        :param default: The default value for the parameter if specified, otherwise `inspect._empty`.
        :param annotation: The annotation for the parameter if specified, otherwise `inspect._empty`.
        :param kind: Describes how argument values are bound to the parameter.
        :param description: String description of the parameter.
        """
        super().__init__(*args, **kwargs)
        self.description = description

    @classmethod
    def from_parameter(cls, parameter: inspect.Parameter):
        """
        Creates :class:`ParameterWithDescription` instance from :class:`inspect.Parameter` instance.

        :param parameter: Parameter to convert into :class:`ParameterWithDescription`
        """
        return cls(
            name=parameter.name,
            kind=parameter.kind,
            default=parameter.default,
            annotation=parameter.annotation,
        )

    def to_parameter(self) -> inspect.Parameter:
        """
        Converts this instance of :class:`ParameterWithDescription` into a :class:`inspect.Parameter` instance.
        """
        return inspect.Parameter(
            name=self.name,
            kind=self.kind,
            default=self.default,
            annotation=self.annotation,
        )

    def __eq__(self, other):
        if super().__eq__(other) and isinstance(other, ParameterWithDescription):
            return self.description == other.description
        return False

    def __hash__(self):
        return hash((super().__hash__(), self.description))

    def replace(self, *, description=inspect._empty, **kwargs):
        """
        Generates a new instance and changes some of the properties.

        :param name: The name of the parameter as a string.
        :param default: The default value for the parameter if specified, otherwise `inspect._empty`.
        :param annotation: The annotation for the parameter if specified, otherwise `inspect._empty`.
        :param kind: Describes how argument values are bound to the parameter.
        :param description: String description of the parameter.
        """
        value = super().replace(**kwargs)
        value.description = self.description
        assert isinstance(value, ParameterWithDescription)
        if description is not inspect._empty:
            value.description = description
        return value


class SignatureWithDescription(inspect.Signature):
    """
    Extends `inspect.Signature` class with short and long description parameters.

    :attr params: Signature parameters.
    :attr return_annotation: Type annotation of the value returned by the function.
    :attr short_description: Short description of the function.
    :attr long_description: Long description of the function.
    """

    def __init__(
        self,
        *args,
        short_description: t.Optional[str] = None,
        long_description: t.Optional[str] = None,
        **kwargs,
    ):
        """
        Extends `inspect.Signature` class with short and long description parameters.

        :param params: Signature parameters.
        :param return_annotation: Type annotation of the value returned by the function.
        :param short_description: Short description of the function.
        :param long_description: Long description of the function.
        """
        super().__init__(*args, **kwargs)
        self.short_description = short_description
        self.long_description = long_description

    @classmethod
    def from_callable(cls, obj: t.Callable, **kwargs) -> "SignatureWithDescription":
        """
        Generates the signature from a callable object.

        :param obj: Callable function to generate signature from
        """
        signature = super(cls, cls).from_callable(obj, **kwargs)
        parameters = []

        # Add help
        doc = docstring_parser.parse(obj.__doc__ or "")
        doc_params = {p.arg_name: p.description for p in doc.params}
        for p in signature.parameters.values():
            if not isinstance(p, ParameterWithDescription):
                p = t.cast(
                    ParameterWithDescription, ParameterWithDescription.from_parameter(p)
                )
                p.description = doc_params.get(p.name)
            parameters.append(p)
        signature = SignatureWithDescription(
            parameters=parameters, return_annotation=signature.return_annotation
        )
        signature.long_description = doc.long_description
        signature.short_description = doc.short_description
        return signature

    def __eq__(self, other):
        if super().__eq__(other) and isinstance(other, SignatureWithDescription):
            return (
                self.short_description == other.short_description
                and self.long_description == other.long_description
            )
        return False

    def __hash__(self):
        return hash((super().__hash__(), self.long_description, self.short_description))

    def replace(
        self,
        *,
        short_description=inspect._empty,
        long_description=inspect._empty,
        **kwargs,
    ):
        """
        Generates a new instance and changes some of the properties.

        :param params: Signature parameters.
        :param return_annotation: Type annotation of the value returned by the function.
        :param short_description: Short description of the function.
        :param long_description: Long description of the function.
        """
        value = super().replace(**kwargs)
        assert isinstance(value, SignatureWithDescription)
        value.short_description = self.short_description
        value.long_description = self.long_description
        if short_description is not inspect._empty:
            value.short_description = short_description
        if long_description is not inspect._empty:
            value.long_description = long_description
        return value


def signature_with_description(obj: t.Callable, **kwargs) -> SignatureWithDescription:
    """
    Takes a signature of a callable and parses its docstring to
    add its description and the description of all its parameters.

    :param obj: Callable function to generate signature from
    """
    return SignatureWithDescription.from_callable(obj, **kwargs)


def copy_signature(
    other_function: t.Callable, copy_return_annotation: bool = False
) -> t.Callable[[T], T]:
    """This is a decorator used to copy signature from another function
    called inside this function. The parameters from the called function
    are added only if current function is able to receive them (it has
    either `*args`, or `**kwargs` argument. The signature is copied including
    the description of all parameters.

    Example usage::

        def a(param1: int):
            pass

        @copy_signature(a)
        def b(param2, *args, **kwargs):
            a(*args, **kwargs)

    :param other_function: Signature to extend the current function with (i.e. the called function)
    :param copy_return_annotation: Whether the return type annotation should be taken from the called function.
    """

    def wrap(fn: T) -> T:
        other_signature = signature_with_description(other_function)
        current_signature = signature_with_description(fn)
        signature = _merge_signatures(
            current_signature,
            other_signature,
            copy_return_annotation=copy_return_annotation,
        )
        setattr(fn, "__signature__", signature)
        return fn

    return wrap


def get_class_name(class_type: t.Type) -> str:
    """
    Returns customizable name of the class in a human-readable format.
    By default a name in the PascalCase is converted to the snake_case.

    :param class_type: Class to take the name from.
    """
    if hasattr(class_type, "_get_class_name"):
        return class_type._get_class_name()
    name = class_type.__name__
    name = _pascal_to_snake_case(name)
    return name


@t.overload
def default_from_str() -> t.Callable[[TType], TType]:
    ...  # pragma: no cover


@t.overload
def default_from_str(cls: TType) -> TType:
    ...  # pragma: no cover


def default_from_str(cls=None):
    """A decorator that extends the class with the :func:`from_str`, :func:`__str__`, and :func:`__str_with_dashes_option__`
    methods. The :func:`from_str` method is a classmethod and it parses a string value into
    the instance of the class. If some class derive this class, this class will be able to parse
    the string value as these instances as well. In this case, :func:`from_str` should be called on this
    type (parent class).

    :func:`__str__` and :func:`__str_with_dashes_option__` are implemented as inverse of :func:`from_str` (i.e., the
    strings generated by these methods can be parsed using :func:`from_str` method). :func:`__str_with_dashes_option__`
    further allows to specify if the underscores in class names and property names should be converted to dashes.

    Example usage::

        @default_from_str
        class A:
            def __init__(self, prop1: int):
                self.prop1 = prop1

        str(A(42)) == "a(prop1=42)"

    """

    def decorator(cls: TType) -> TType:
        _validate_class_supports_to_str(cls)
        setattr(
            cls,
            "from_str",
            classmethod(
                partial(
                    _parse_class_structure_for_all_descendents,
                    base_cls=cls,
                    method_name="from_str",
                )
            ),
        )
        setattr(cls, "__str_with_dashes_option__", _class_to_str_with_dashes_option)
        setattr(cls, "__str__", _class_to_str)
        return cls

    if cls is not None:
        return decorator(cls)
    return decorator


def parse_class_structure(
    value: str,
    classes: t.List[t.Type],
    known_classes: t.Optional[t.List[t.Type]] = None,
) -> t.Any:
    """Parses a string representing a structure of classes into a
    tree representation of the classes. This method is directly
    used by the :func:`default_from_str` decorator.

    The following conventions are used:

    - Classes are represented as `class_name(arg1, arg2, named_arg1=val3, ...)`
    - Lists are represented as `[val1, val2, ...]`
    - Tuples are represented as `(val1, val2, ...)`
    - Dicts are represented as `{key1=val1, key2=val2, ...}`
    - Strings can be represented as `value` or `"complicated 'value'"` or `'complicated "value"'`
    - Numbers and bool are represented as strings: `15`, `16.2`, `true`

    :param value: String value to parse.
    :param classes: List of classes to parse the string into.
    :param known_classes: List of known classes that can be used during parsing on
                          lower levels (e.g. as a property value).
    :return: An instance of one of the classes is returned.
    """
    known_classes = _find_known_classes(classes)
    known_classes_map = {get_class_name(x): x for x in chain(classes, known_classes)}
    class_map = {get_class_name(x): x for x in classes}

    def _map_class(class_argument, value_type, is_root=True):
        if value_type is not None and hasattr(value_type, "from_str"):
            return value_type.from_str(str(class_argument))
        elif isinstance(class_argument, _ClassArgument):
            known_classes = class_map if is_root else known_classes_map
            if class_argument.name not in known_classes:
                raise RuntimeError(
                    f'Could not find class with name "{class_argument.name}" in the list of registered classes: {", ".join(sorted(known_classes.keys()))}'
                )
            class_type = known_classes[class_argument.name]
            signature = _full_signature(class_type)
            args_types, kwargs_types = _get_signature_args_kwargs(signature)
            num_pos_args_required = sum(
                1
                for p in signature.parameters.values()
                if p.kind == inspect.Parameter.POSITIONAL_ONLY
                and p.default is inspect._empty
            )
            required_keys = {
                p.name
                for p in signature.parameters.values()
                if p.kind
                in {
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.KEYWORD_ONLY,
                }
                and p.default is inspect._empty
            }
            if len(class_argument.args) > len(args_types):
                raise RuntimeError(
                    f"Number of passed positional arguments ({len(class_argument.args)}) to class {class_argument.name} exceeds the number of allowed arguments ({len(args_types)})."
                )
            elif len(class_argument.args) < num_pos_args_required:
                raise RuntimeError(
                    f"Number of passed positional arguments ({len(class_argument.args)}) to class {class_argument.name} is lower then the number of expected arguments ({num_pos_args_required})."
                )
            unknown_keys = set(class_argument.kwargs.keys()).difference(
                set(kwargs_types.keys())
            )
            if unknown_keys:
                raise RuntimeError(
                    f"There were unknown parameters {{{', '.join(sorted(unknown_keys))}}} to class {class_argument.name}. Allowed parameters are: {{{', '.join(sorted(kwargs_types.keys()))}}}"
                )
            missing_keys = required_keys.difference(class_argument.kwargs.keys())
            missing_keys.difference_update(
                x for x, _ in args_types[: len(class_argument.args)]
            )
            if missing_keys:
                raise RuntimeError(
                    f'Parameters {{{", ".join(sorted(missing_keys))}}} to class {class_argument.name} are missing. Passed arguments are: {{{", ".join(sorted(class_argument.kwargs.keys()))}}}.'
                )
            args = [
                _map_class(x, tp, is_root=False)
                for x, (_, tp) in zip(class_argument.args, args_types)
            ]
            kwargs = OrderedDict(
                (k, _map_class(v, kwargs_types[k], is_root=False))
                for k, v in class_argument.kwargs.items()
            )
            return class_type(*args, **kwargs)

        assert value_type is not None, "Type of value must be supplied"
        if getattr(value_type, "__origin__", None) is t.Union:
            # Add Optional support
            if sum(1 for x in value_type.__args__ if x not in {type(None)}) == 1:
                value_type = next(
                    x for x in value_type.__args__ if x not in {type(None)}
                )
                if str(class_argument).lower() in {"none", "null", ""}:
                    return None

        if getattr(value_type, "__origin__", None) in (list,):
            class_argument = [
                _map_class(x, value_type.__args__[0], is_root=False)
                for x in class_argument
            ]
            value_type = value_type.__origin__
        if getattr(value_type, "__origin__", None) in (Literal,):
            class_argument = _map_class(class_argument, str, is_root=False)
            if class_argument not in value_type.__args__:
                raise RuntimeError(
                    f"Literal value {class_argument} is not in the set of supported values {{{', '.join(sorted(value_type.__args__))}}}"
                )
            value_type = str
        if getattr(value_type, "__origin__", None) in (tuple,):
            if len(class_argument) != len(value_type.__args__):
                raise RuntimeError(
                    f"Cannot parse {len(class_argument)} values as tuple of length {len(value_type.__args__)}"
                )
            class_argument = [
                _map_class(x, tp, is_root=False)
                for tp, x in zip(value_type.__args__, class_argument)
            ]
            value_type = value_type.__origin__

        if getattr(value_type, "__origin__", None) in (dict, OrderedDict):
            if not isinstance(class_argument, OrderedDict):
                raise RuntimeError(
                    f"Cannot parse {class_argument} as a {value_type.__origin__.__name__} instance"
                )
            class_argument = [
                (
                    _map_class(k, value_type.__args__[0], is_root=False),
                    _map_class(v, value_type.__args__[1], is_root=False),
                )
                for k, v in class_argument.items()
            ]
            value_type = value_type.__origin__

        if value_type == bool:
            if not isinstance(class_argument, str):
                raise RuntimeError(f"Cannot parse {class_argument} as bool")

            norm = class_argument.strip().lower()
            if norm in {"1", "true", "t", "yes", "y", "on"}:
                return True

            if norm in {"0", "false", "f", "no", "n", "off"}:
                return False

            raise RuntimeError(f"Cannot parse {class_argument} as bool")
        return value_type(class_argument)

    assert isinstance(value, str)
    class_tree = _ClassArgument.from_str(value)
    value = _map_class(class_tree, None)
    return value


TSignature = t.TypeVar("TSignature", bound=inspect.Signature)


def _merge_signatures(
    current_signature: TSignature,
    *signatures: TSignature,
    copy_return_annotation: bool = False,
) -> TSignature:
    if len(signatures) == 0:
        return current_signature

    other_signature = signatures[0]
    parameters = []
    has_args = False
    for p in current_signature.parameters.values():
        if p.kind in {
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        }:
            parameters.append(p)
        if p.kind == inspect.Parameter.VAR_POSITIONAL:
            for p2 in other_signature.parameters.values():
                if p2.kind in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.VAR_POSITIONAL,
                }:
                    parameters.append(p2)
            has_args = True
        if p.kind == inspect.Parameter.VAR_KEYWORD:
            for p2 in other_signature.parameters.values():
                if p2.kind in {
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.VAR_KEYWORD,
                }:
                    parameters.append(p2)
                elif (
                    p2.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD and not has_args
                ):
                    parameters.append(
                        p2.replace(
                            kind=inspect.Parameter.KEYWORD_ONLY,
                        )
                    )
    return_annotation = (
        other_signature.return_annotation
        if copy_return_annotation
        else current_signature.return_annotation
    )
    signature = current_signature.replace(
        parameters=parameters, return_annotation=return_annotation
    )
    if len(signatures) > 1:
        signature = _merge_signatures(
            signature, *signatures[1:], copy_return_annotation=copy_return_annotation
        )
    return signature


def _full_signature(fn: t.Callable) -> SignatureWithDescription:
    assert callable(fn)

    if inspect.isclass(fn):
        signature = _merge_signatures(
            *[signature_with_description(x) for x in fn.__mro__]
        )
    else:
        signature = signature_with_description(fn)
    return signature


class _ClassArgument:
    def __init__(self, name, args, kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        args = [
            self.escape_string(v) if isinstance(v, str) else repr(v) for v in self.args
        ] + [
            f"{k}={self.escape_string(v) if isinstance(v, str) else repr(v)}"
            for k, v in self.kwargs.items()
        ]
        return f'{self.name}({", ".join(args)})'

    def __str__(self):
        return repr(self)

    class ParseError(ValueError):
        def __init__(self, message, string, span):
            string = (
                string[: span[0]]
                + "\033[4;31m"
                + string[span[0] : span[1]]
                + "\033[0;0m"
                + string[span[1] :]
            )
            super().__init__(message + "\n" + f'Error in string "{string}" {span}')

    @staticmethod
    def match_string(string, span):
        local_string = string[slice(*span)]
        string_match = re.match(
            r"^(?:(['\"])(?:(?!\1)[^\\]|\\\\|\\\1)*\1|[^\),\]\}]*)", local_string
        )
        assert string_match, f'Could not parse string "{local_string}"'
        return tuple(span[0] + x for x in string_match.span())

    @staticmethod
    def unescape_string(string):
        if len(string) == 0 or string[0] not in {'"', "'"}:
            return string
        assert len(string) > 1
        quotation_mark = string[0]
        string = string[1:-1]
        string = string.replace(f"\\{quotation_mark}", quotation_mark).replace(
            "\\\\", "\\"
        )
        return string

    @staticmethod
    def escape_string(string: str) -> str:
        if {'"', ")", "]", "}", ",", " ", "\r", "\n", "\t"}.intersection(string):
            string = string.replace("\\", "\\\\").replace('"', '\\"')
            return '"' + string + '"'
        return string

    @staticmethod
    def parse_argument(string, span=None):
        if span is None:
            span = 0, len(string)
        out_name = None
        name_match = re.match(r"^([a-zA-Z0-9_-]+)\s*=\s*", string[slice(*span)])
        if name_match:
            out_name = name_match.group(1).replace("-", "_")
            span = (span[0] + name_match.end(), span[1])
        arguments = []
        kwargs = OrderedDict()

        class_match = re.match(r"^([a-zA-Z0-9_-]+)\(", string[slice(*span)])
        if class_match:
            # This is a nested class argument
            end = class_match.end() + span[0]
            if len(string) > end and string[end] != ")":
                while True:
                    (_, new_end), name, val = _ClassArgument.parse_argument(
                        string, span=(end, span[1])
                    )
                    if name is None:
                        arguments.append(val)
                    else:
                        name = name.replace("-", "_")
                        kwargs[name] = val
                    end = new_end
                    par_match = re.match(r"^\s*(,|\))\s*", string[end : span[1]])
                    if par_match:
                        end += par_match.end()
                        if par_match.group(0).strip() == ")":
                            break
                    else:
                        raise _ClassArgument.ParseError(
                            "Unexpected end of argument", string, (end, span[1])
                        )
            elif len(string) > end and string[end] in {")"}:
                end += 1
            else:
                raise _ClassArgument.ParseError(
                    "Unexpected end of argument", string, (end, span[1])
                )
            span = (span[0], end)
            class_name = class_match.group(1).replace("-", "_")
            return (
                span,
                out_name,
                _ClassArgument(class_name, arguments, kwargs),
            )

        list_dict_match = re.match(r"^(\(|\[|\{)\s*", string[slice(*span)])
        if list_dict_match:
            # This is a list argument
            end = list_dict_match.end() + span[0]
            is_dict = list_dict_match.group(1) == "{"
            arguments = []
            if len(string) > end and string[end] not in {")", "}", "]"}:
                while True:
                    (_, new_end), name, val = _ClassArgument.parse_argument(
                        string, span=(end, span[1])
                    )
                    if is_dict:
                        if name is None:
                            raise _ClassArgument.ParseError(
                                "Missing dict entry name", string, (end, new_end)
                            )
                        arguments.append((name, val))
                    elif not is_dict:
                        if name is not None:
                            name, val = None, _ClassArgument.unescape_string(
                                string[end:new_end]
                            )
                        arguments.append(val)

                    end = new_end
                    par_match = re.match(r"^\s*(,|\)|\]|\})\s*", string[end : span[1]])
                    if par_match:
                        end += par_match.end()
                        if par_match.group(0).strip() in {")", "}", "]"}:
                            break
                    if not par_match or end == span[1]:
                        raise _ClassArgument.ParseError(
                            "Unexpected end of argument", string, (end, span[1])
                        )
            elif len(string) > end and string[end] in {")", "}", "]"}:
                end += 1
            else:
                raise _ClassArgument.ParseError(
                    "Unexpected end of argument", string, (end, span[1])
                )

            span = (span[0], end)
            out_type = {"{": OrderedDict, "[": list, "(": tuple}[
                list_dict_match.group(1)
            ]

            return (
                span,
                out_name,
                out_type(arguments),
            )

        # Else:
        span = _ClassArgument.match_string(string, span)
        argument_value = string[slice(*span)]
        argument_value = _ClassArgument.unescape_string(argument_value)
        return span, out_name, argument_value

    @staticmethod
    def from_str(string):
        return _ClassArgument.parse_argument(string)[-1]


def _parse_class_structure_for_all_descendents(
    cls, value: str, *, base_cls, method_name="from_str"
):
    def add_classes_rec(cls):
        yield cls
        for c in cls.__subclasses__():
            yield from add_classes_rec(c)

    classes = list(add_classes_rec(base_cls))
    if base_cls != cls:
        raise RuntimeError(
            f"Method {method_name} should be called on {base_cls}, not {cls}"
        )
    return parse_class_structure(value, classes)


def _validate_class_supports_to_str(cls):
    original_cls = cls

    def inner(cls, ignore_from_str=False):
        if not ignore_from_str and hasattr(cls, "from_str"):
            pass  # from_str objects are supported
        elif getattr(cls, "__origin__", None) == Literal:
            if not all(isinstance(x, str) for x in cls.__args__):
                raise ValueError(
                    f"Class {original_cls} is not supported because it contains a Literal property of non-str types"
                )
        elif getattr(cls, "__origin__", None) in {
            t.Union,
            tuple,
            list,
            set,
            dict,
            OrderedDict,
        }:
            for tp in cls.__args__:
                inner(tp)
            if getattr(cls, "__origin__", None) in (dict, OrderedDict):
                if cls.__args__[0] not in (str,):
                    raise RuntimeError(
                        f"Class {original_cls} is not supported because it contains a dictionary with key of type {cls.__args__[0]}. Only str type is supported"
                    )
        elif cls in (type(None), str, bool, float, int, bytes):
            pass
        elif dataclasses.is_dataclass(cls):
            # Dataclasses are supported
            for f in dataclasses.fields(cls):
                inner(f.type)
        elif _is_class(cls):
            # Classes are supported
            # but they must set all constructor parameters as properties
            # this is verified on runtime
            args_types, kwargs_types = _get_signature_args_kwargs(cls, exclusive=True)
            for p_name, p_type in chain(args_types, kwargs_types.items()):
                if p_type is inspect._empty:
                    raise ValueError(
                        f"Class {original_cls} is not supported because field {p_name} in {cls} does not have a type annotation"
                    )
                inner(p_type)
        else:
            raise ValueError(
                f"Class {original_cls} is not supported because it contains type {cls} which is not supported"
            )

    inner(cls, ignore_from_str=True)


def _class_to_str_with_dashes_option(self, use_dashes=False) -> str:
    def inner(obj, ignore_from_str=True, use_dashes=False) -> str:
        if not ignore_from_str and hasattr(obj, "from_str"):
            if hasattr(obj, "__str_with_dashes_option__"):
                return _ClassArgument.escape_string(
                    obj.__str_with_dashes_option__(use_dashes=use_dashes)
                )
            else:
                return _ClassArgument.escape_string(str(obj))
        elif isinstance(obj, tuple):
            return (
                "("
                + ", ".join(inner(x, False, use_dashes=use_dashes) for x in obj)
                + ")"
            )
        elif isinstance(obj, (list, set)):
            return (
                "["
                + ", ".join(inner(x, False, use_dashes=use_dashes) for x in obj)
                + "]"
            )
        elif isinstance(obj, (dict, OrderedDict)):
            return (
                "{"
                + ", ".join(
                    str(k) + "=" + inner(x, False, use_dashes=use_dashes)
                    for k, x in obj.items()
                )
                + "}"
            )
        elif isinstance(obj, (bool, float, int, bytes, type(None))):
            return str(obj)
        elif isinstance(obj, str):
            return _ClassArgument.escape_string(obj)
        elif dataclasses.is_dataclass(obj):
            class_name = get_class_name(type(obj))
            args: t.List[str] = []
            kwargs = {}
            for f in dataclasses.fields(obj):
                name = f.name
                if use_dashes:
                    name = name.replace("_", "-")
                kwargs[name] = inner(getattr(obj, f.name), False, use_dashes=use_dashes)
                # if getattr(f, 'kw_only', False):
                # else:
                #     args.append(getattr(self, f.name))
            args.extend((f"{k}={v}" for k, v in kwargs.items()))
            if use_dashes:
                class_name = class_name.replace("_", "-")
            return f'{class_name}({", ".join(args)})'
        elif _is_class(type(obj)):
            class_name = get_class_name(type(obj))
            args_par, kwargs_par = _get_signature_args_kwargs(type(obj), exclusive=True)
            args = []
            kwargs = {}
            for p_name, _ in args_par:
                if not hasattr(obj, p_name):
                    raise RuntimeError(
                        f"Cannot serialize class {type(obj)}, because the constructor parameter {p_name} is missing from class's properties"
                    )
                value = getattr(obj, p_name)
                args.append(inner(value, False, use_dashes=use_dashes))
            for name, p in kwargs_par.items():
                if not hasattr(obj, name):
                    raise RuntimeError(
                        f"Cannot serialize class {type(obj)}, because the constructor parameter {name} is missing from class's properties"
                    )
                value = getattr(obj, name)
                if use_dashes:
                    name = name.replace("_", "-")
                args.append(name + "=" + inner(value, False, use_dashes=use_dashes))
            if use_dashes:
                class_name = class_name.replace("_", "-")
            return f'{class_name}({", ".join(args)})'
        else:
            raise RuntimeError(f"Type {type(obj)} is not supported")

    return inner(self, use_dashes=use_dashes)


def _class_to_str(self):
    return self.__str_with_dashes_option__(use_dashes=False)


def _is_class(tp):
    if not inspect.isclass(tp):
        return False

    if tp in (type(None), types.FunctionType):
        return False

    if tp in vars(builtins).values():
        return False

    return True


def _pascal_to_snake_case(val):
    out = ""
    for x in val:
        if x.isupper():
            if len(out) > 0:
                out += "_"
            out += x.lower()
        else:
            out += x
    return "".join(out)


def _find_known_classes(classes) -> t.List[t.Type]:
    known_classes = []
    classes = deque(classes)
    while classes:
        c = classes.pop()
        known_classes.append(c)
        for p in _full_signature(c).parameters.values():
            if getattr(p.annotation, "__origin__", None) in {
                list,
                set,
                tuple,
                dict,
                t.Union,
                OrderedDict,
            }:
                classes.extend(x for x in p.annotation.__args__ if _is_class(x))
            elif _is_class(p.annotation):
                classes.append(p.annotation)
    return known_classes


def _get_help_text(signature) -> t.Tuple[str, str]:
    vals = [x for x in (signature.short_description, signature.long_description) if x]
    return (
        inspect.cleandoc(signature.short_description or ""),
        inspect.cleandoc("\n\n".join(vals)),
    )


def _get_signature_args_kwargs(signature, exclusive=False):
    if not isinstance(signature, inspect.Signature):
        signature = _full_signature(signature)
    signature = list(signature.parameters.values())
    args = [
        (p.name, p.annotation)
        for p in signature
        if p.kind
        in (
            {
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            }
            if not exclusive
            else {
                inspect.Parameter.POSITIONAL_ONLY,
            }
        )
    ]
    kwargs = OrderedDict(
        (p.name, p.annotation)
        for p in signature
        if p.kind
        in {inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD}
    )
    return args, kwargs


def _wrap_fn_to_allow_kwargs_instead_of_args(fn):
    signature = _full_signature(fn)
    args_params = [
        p
        for p in signature.parameters.values()
        if p.kind in {inspect.Parameter.POSITIONAL_ONLY}
    ]
    kwargs_params = [
        p
        for p in signature.parameters.values()
        if p.kind
        in {inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD}
    ]
    if len(args_params) == 0:
        return fn

    def inner_fn(**kwargs):
        args = []
        for p in args_params:
            if p.name not in kwargs:
                break
            args.append(kwargs.pop(p.name))
        return fn(*args, **kwargs)

    parameters = []
    for p in args_params:
        parameters.append(
            p.replace(
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            )
        )
    parameters.extend(kwargs_params)
    setattr(
        inner_fn,
        "__signature__",
        signature.replace(
            parameters=parameters, return_annotation=signature.return_annotation
        ),
    )
    setattr(inner_fn, "__doc__", fn.__doc__)
    return inner_fn



import typing as t

import click as _click

from .core import Command, Group
from .utils import _full_signature, _get_help_text


CmdType = t.TypeVar("CmdType", bound=Command)
ClickCmdType = t.TypeVar("ClickCmdType", bound=_click.Command)
F = t.TypeVar("F", bound=t.Callable[..., t.Any])


@t.overload
def command(
    __func: t.Callable[..., t.Any],
) -> Command:
    ...  # pragma: no cover


@t.overload
def command(
    name: t.Optional[str] = None,
    **attrs: t.Any,
) -> t.Callable[..., Command]:
    ...  # pragma: no cover


@t.overload
def command(
    name: t.Optional[str] = None,
    cls: t.Type[CmdType] = ...,
    **attrs: t.Any,
) -> t.Callable[..., CmdType]:
    ...  # pragma: no cover


@t.overload
def command(
    name: t.Optional[str] = None,
    cls: t.Type[ClickCmdType] = ...,
    **attrs: t.Any,
) -> t.Callable[..., ClickCmdType]:
    ...  # pragma: no cover


def command(
    name: t.Union[str, t.Callable, None] = None,
    cls: t.Optional[t.Type[_click.Command]] = None,
    **attrs: t.Any,
) -> t.Union[_click.Command, t.Callable[..., _click.Command]]:
    r"""Creates a new :class:`Command` and uses the decorated function as
    callback. This will automatically parse the function signature (if
    no other signature is supplied) and generates `Option` and `Argument`
    instance accordingly. User can override individual parameters either
    by using `click.option` and `click.argument` decorators or by passing
    custom parameters with the same name as the parameters to override.
    The name of the command defaults to the name of the function with
    underscores replaced by dashes. If you want to change that, you can
    pass the intended name as the first argument.
    All keyword arguments are forwarded to the underlying command class.
    Once decorated the function turns into a :class:`Command` instance
    that can be invoked as a command line utility or be attached to a
    command :class:`Group`.

    NOTE: This decorator should always be used instead of `click` command
    decorator, because in `click`, the help is automatically populated
    with the callback's `__doc__` and it is not parsed correctly.

    :param name: the name of the command.  This defaults to the function
                 name with underscores replaced by dashes.
    :param cls: the command class to instantiate.  This defaults to
                :class:`Command`. If the class is not a subclass of :class:`Command`,
                it is automatically wrapped.
    :param hierarchical: whether to parse complex types using hierarchical
                         parsing instead of inline parsing. In hierarchical
                         parsing the complex class options are expanded into
                         individual options corresponding to individual
                         properties. Note that some types are not supported
                         with hierarchical parsing. Default is to use inline
                         parsing.
    :param map_parameter_name: a function that maps from parameter paths
                               (e.g. param1.property1.subproperty2) into
                               parameter name used in parsing.
    """
    func: t.Optional[t.Callable] = None

    if callable(name):
        func = name
        name = None
        assert cls is None, "Use 'command(cls=cls)(callable)' to specify a class."
        assert not attrs, "Use 'command(**kwargs)(callable)' to provide arguments."

    command_name: t.Optional[str] = name

    if cls is None:
        cls = Command
    elif not issubclass(cls, Command):
        cls = type(f"{cls.__name__}AClickWrapper", (Command, cls), {})

    def decorator(fn):
        if attrs.get("help") is None or attrs.get("short_help") is None:
            short_help, help = _get_help_text(_full_signature(fn))
            if attrs.get("help") is None:
                attrs["help"] = help
            if attrs.get("short_help") is None:
                attrs["short_help"] = short_help
        click_decorator = _click.command(name=command_name, cls=cls, **attrs)
        return click_decorator(fn)

    if func is not None:
        return decorator(func)

    return decorator


@t.overload
def group(
    __func: t.Callable[..., t.Any],
) -> Group:
    ...  # pragma: no cover


@t.overload
def group(
    name: t.Optional[str] = None,
    **attrs: t.Any,
) -> t.Callable[[F], Group]:
    ...  # pragma: no cover


def group(
    name: t.Union[str, t.Callable[..., t.Any], None] = None, **attrs: t.Any
) -> t.Union[Group, t.Callable[[F], Group]]:
    """
    Creates a new :class:`Group` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.
    """
    if attrs.get("cls") is None:
        attrs["cls"] = Group

    if callable(name):
        grp: t.Callable[[F], Group] = t.cast(Group, command(**attrs))
        return grp(name)

    return t.cast(Group, command(name, **attrs))
